import asyncio
import io
import logging
import os.path
from dataclasses import dataclass
from typing import (
    Iterable,
    List,
    Literal,
    Optional,
    Tuple,
)

import confuse
import pypdf
import yaml
from aiokit import AioThing
from izihawa_utils.exceptions import BaseError
from izihawa_utils.file import mkdir_p
from stc_geck.advices import BaseDocumentHolder
from stc_geck.client import StcGeck

from .chains.map_reduce import (
    QAChain,
    SummarizeChain,
)
from .configs import ConfigGenerator
from .data_source.base import SourceDocument
from .data_source.geck_data_source import GeckDataSource
from .document_chunker import (
    Chunk,
    DocumentChunker,
)
from .model import KadiaModel
from aiokit import MultipleAsyncExecution
from .vector_storage.qdrant import (
    QdrantVectorStorage,
    ScoredChunk,
)


class DocumentNotFoundError(BaseError):
    pass


def print_color(text, color):
    print("\033[38;5;{}m{}\033[0m".format(color, text))


@dataclass
class KadiaResponse:
    answer: str
    chunks: List[Chunk]


class KadiaAI(AioThing):
    def __init__(
        self,
        home_path: Optional[str] = None,
        geck: Optional[StcGeck] = None,
    ):
        """
        Main Kadia class that manages AI operations

        :param home_path: path to config and/or embeddings directory
        :param geck: an instance of GECK
        """
        super().__init__()
        self.home_path = self.get_home_path(home_path)
        config = confuse.Configuration("kadia", __name__)
        config_path = self.ensure_config()
        config.set_file(config_path)

        self.model = KadiaModel(config["model"])
        self.document_chunker = DocumentChunker(
            text_splitter=self.model.text_splitter,
            add_metadata=self.model.config["text_splitter"]["add_metadata"].get(bool),
        )

        self.geck = geck
        if not self.geck:
            self.geck = StcGeck(
                ipfs_http_base_url=config["ipfs"]["http"]["base_url"].get(str),
                grpc_api_endpoint=config["summa"]["endpoint"].get(str),
                timeout=600,
            )
            self.starts.append(self.geck)

        self.data_source = GeckDataSource(self.geck)
        self.vector_storage = QdrantVectorStorage(
            qdrant_config=config["qdrant"],
            collection_name=self.model.get_embeddings_id(),
            embedding_function=self.model.embed_documents,
            force_recreate=config["qdrant"]["force_recreate"].get(bool),
        )

    async def _get_missing_chunks_job(self, all_chunks, document):
        document_chunks = await self.generate_chunks_from_document(document)
        all_chunks.extend(document_chunks)

    async def _get_missing_chunks(
        self,
        documents: List[SourceDocument],
        skip_downloading_pdf: bool = True,
        par: int = 16,
    ) -> List[Chunk]:
        all_chunks = []
        executor = MultipleAsyncExecution(par)
        for document in documents:
            is_stored = await asyncio.get_running_loop().run_in_executor(
                None,
                lambda document_id: self.vector_storage.exists_by_field_value(
                    "document_id", document_id
                ),
                document.document_id,
            )
            if is_stored:
                logging.getLogger("statbox").info(
                    {
                        "action": "already_stored",
                        "mode": "kadia",
                        "document_id": document.document_id,
                    }
                )
                continue
            if skip_downloading_pdf and "content" not in document.document:
                logging.getLogger("statbox").info(
                    {
                        "action": "no_content",
                        "mode": "kadia",
                        "document_id": document.document_id,
                    }
                )
                continue
            try:
                await executor.execute(
                    self._get_missing_chunks_job(all_chunks, document)
                )
            except ValueError:
                logging.getLogger("statbox").info(
                    {
                        "action": "broken_content",
                        "mode": "kadia",
                        "document_id": document.document_id,
                    }
                )
        await executor.join()
        return all_chunks

    async def _search_in_vector_storage(
        self,
        query: str,
        n_chunks: int = 3,
        field_values: Optional[Iterable[Tuple[str, str]]] = None,
        minimum_score: float = 0.5,
    ) -> List[ScoredChunk]:
        logging.getLogger("statbox").info(
            {
                "action": "query",
                "mode": "kadia",
                "query": query,
                "n_chunks": n_chunks,
                "field_values": field_values,
            }
        )
        chunks = await asyncio.get_running_loop().run_in_executor(
            None,
            lambda: self.vector_storage.query(
                self.model.embedder.embed_query(query),
                n_chunks=n_chunks,
                field_values=field_values,
            ),
        )
        filtered_chunks = []
        for chunk in chunks:
            if chunk.score > minimum_score:
                filtered_chunks.append(chunk)
        logging.getLogger("statbox").info(
            {
                "action": "query",
                "mode": "kadia",
                "found": len(filtered_chunks),
                "minimum_score": minimum_score,
            }
        )
        return filtered_chunks

    def get_home_path(self, home_path: str) -> str:
        """
        Expands path to config and/or embeddings directory and ensures the directory existence

        :param home_path:
        :return:
        """
        if home_path is None:
            home_path = os.environ.get("KADIA_HOME", "~/.kadia")
        home_path = os.path.expanduser(home_path)
        if not os.path.exists(home_path):
            mkdir_p(home_path)
        return home_path

    def ensure_config(
        self,
        ipfs_http_base_url: str = "http://127.0.0.1:8080",
        summa_endpoint: str = "127.0.0.1:10082",
        qdrant_base_url: str = "http://127.0.0.1",
        llm_name: Literal[
            "llama-2-7b",
            "llama-2-7b-uncensored",
            "llama-2-13b",
            "openai",
            "petals-llama-2-70b",
            "petals-stable-beluga",
            "mistral-7b",
        ] = "mistral-7b",
        embedder_name: Literal[
            "instructor-xl", "openai", "bge-small-en"
        ] = "bge-small-en",
        device: str = "cpu",
        gpu_layers: int = 50,
        force: bool = False,
    ):
        """
        Write config to $KADIA_HOME/config.yaml
        :param ipfs_http_base_url: IPFS HTTP base url, i.e. `http://127.0.0.1:8080`
        :param summa_endpoint: Summa endpoint, i.e. `127.0.0.1:10082`
        :param qdrant_base_url:
        :param llm_name: 'llama-2-7b', 'llama-2-7b-uncensored', 'llama-2-13b', 'openai', 'petals-llama-2-70b', 'petals-stable-beluga', 'mistral-7b'
        :param embedder_name: 'instructor-xl', 'openai', 'bge-small-en''
        :param device: 'cpu' or 'cuda'
        :param gpu_layers: number of layers to enabled offloading part of calculations to GPU
        :param force: overwrite even if config already exists
        :return:
        """
        config_path = os.path.join(self.home_path, "config.yaml")
        if not os.path.exists(config_path) or force:
            config = {
                "ipfs": {
                    "http": {
                        "base_url": ipfs_http_base_url,
                    }
                },
                "model": ConfigGenerator.default_config(
                    llm_name=llm_name,
                    embedder_name=embedder_name,
                    device=device,
                    gpu_layers=gpu_layers,
                ),
                "qdrant": {
                    "url": qdrant_base_url,
                    "prefer_grpc": True,
                },
                "summa": {
                    "endpoint": summa_endpoint,
                },
            }
            with open(config_path, "w") as f:
                f.write(yaml.dump(config, default_flow_style=False))
        return config_path

    async def resolve_document_content(self, document: SourceDocument) -> Optional[str]:
        """
        Retrieves document content from `content` field or from underlying PDF file.

        :param document:
        :return:
        """
        document = document.document
        if "content" in document:
            return document["content"]
        document_holder = BaseDocumentHolder(document)
        # ToDo: should also utilize epub links
        if pdf_link := document_holder.get_links().get_link_with_extension("pdf"):
            file_content = await self.geck.download(pdf_link["cid"])
            pdf_reader = pypdf.PdfReader(io.BytesIO(file_content))
            return "\n".join(page.extract_text() for page in pdf_reader.pages)

    async def upsert_documents(
        self, documents: List[SourceDocument], skip_downloading_pdf: bool = True
    ):
        """
        Upsert documents into vector storage

        :param documents:
        :param skip_downloading_pdf:
        :return:
        """

        if not documents:
            return
        chunks = await self._get_missing_chunks(
            documents, skip_downloading_pdf=skip_downloading_pdf
        )
        if chunks:
            logging.getLogger("statbox").info(
                {
                    "action": "add_full_documents",
                    "mode": "kadia",
                    "n": len(chunks),
                }
            )
            await asyncio.get_running_loop().run_in_executor(
                None, lambda: self.vector_storage.upsert(chunks)
            )
            logging.getLogger("statbox").info(
                {
                    "action": "added_full_documents",
                    "mode": "kadia",
                    "n": len(chunks),
                }
            )

    async def upsert_document_by_query(
        self, query: str, skip_downloading_pdf: bool = True
    ):
        """
        Query documents from data source and upsert them into vector storage

        :param query:
        :param skip_downloading_pdf:
        :return:
        """
        documents = await self.data_source.search_documents(query, limit=1)
        if not documents:
            raise DocumentNotFoundError(id_=query)
        await self.upsert_documents(
            documents, skip_downloading_pdf=skip_downloading_pdf
        )
        return documents[0].document_id

    async def search_documents(
        self, query: str, n_documents: int = 10, use_only_keywords: bool = False
    ) -> List[SourceDocument]:
        if not n_documents:
            return []

        # Keywords extraction is required to reduce the number STC access requests (more words => more network requests)
        if use_only_keywords and self.model.keyword_extractor:
            logging.getLogger("statbox").info(
                {
                    "action": "extract_keywords",
                    "mode": "kadia",
                    "query": query,
                }
            )
            keywords = await asyncio.get_running_loop().run_in_executor(
                None,
                lambda: self.model.keyword_extractor.extract_keywords(
                    query,
                    keyphrase_ngram_range=(1, 1),
                ),
            )

            keywords = list(
                map(
                    lambda x: x[1][0],
                    filter(lambda x: x[1][1] > 0.5 or x[0] < 2, enumerate(keywords)),
                )
            )
            query = " ".join(keywords)

        logging.getLogger("statbox").info(
            {
                "action": "query",
                "mode": "kadia",
                "query": query,
            }
        )
        documents = await self.data_source.search_documents(
            query=query, limit=n_documents
        )
        return documents

    async def get_documents_from_chunks(self, chunks):
        """
        Return original documents using GECK and identifiers from chunks

        :param chunks:
        :return:
        """
        ids = set([chunk["document_id"] for chunk in chunks])
        subqueries = []
        for id_ in ids:
            field, value = id_.split(":", 1)
            subqueries.append(
                {"query": {"match": {"value": f'{field}:"{value}"'}}, "occur": "should"}
            )

        search_request = {
            "index_alias": "nexus_science",
            "query": {"boolean": {"subqueries": subqueries}},
            "collectors": [{"top_docs": {"limit": len(subqueries)}}],
            "is_fieldnorms_scoring_enabled": False,
        }
        return await self.geck.get_summa_client().search_documents(search_request)

    async def semantic_search_in_documents(
        self,
        query: str,
        documents: List[SourceDocument],
        n_chunks: int = 10,
        minimum_score: float = 0.5,
        skip_downloading_pdf: bool = True,
    ) -> List[ScoredChunk]:
        await self.upsert_documents(
            documents, skip_downloading_pdf=skip_downloading_pdf
        )
        return await self._search_in_vector_storage(
            query=query,
            n_chunks=n_chunks,
            field_values=[
                ("document_id", document.document_id) for document in documents
            ],
            minimum_score=minimum_score,
        )

    async def semantic_search(
        self,
        query: str,
        n_chunks: int = 10,
        n_documents: int = 30,
        minimum_score: float = 0.5,
        skip_downloading_pdf: bool = True,
        use_only_keywords: bool = True,
    ) -> List[ScoredChunk]:
        """
        Flow for retrieving chunks by chunking documents relevant to `query`

        :param skip_downloading_pdf:
        :param query:
        :param n_chunks:
        :param n_documents:
        :param minimum_score:
        :param use_only_keywords:
        :return:
        """
        documents = await self.search_documents(
            query, n_documents, use_only_keywords=use_only_keywords
        )
        await self.upsert_documents(
            documents, skip_downloading_pdf=skip_downloading_pdf
        )
        return await self._search_in_vector_storage(
            query=query,
            n_chunks=n_chunks,
            minimum_score=minimum_score,
        )

    async def chat_document(
        self, document_id: str, query: str, n_chunks: int, minimum_score: float = 0.5
    ) -> KadiaResponse:
        """
        Flow for getting document by `document_id` and finding answer in this document.

        :param document_id:
        :param query:
        :param n_chunks:
        :param minimum_score:
        :return:
        """

        # Hint: `document_id` is a valid query in STC
        document_id = await self.upsert_document_by_query(
            str(document_id), skip_downloading_pdf=False
        )
        scored_chunks = await self._search_in_vector_storage(
            query=query,
            n_chunks=n_chunks,
            field_values=(("document_id", document_id),),
            minimum_score=minimum_score,
        )
        chunks = [scored_chunk.chunk for scored_chunk in scored_chunks]

        chain = QAChain(query=query, llm_manager=self.model.llm_manager)
        answer = await asyncio.get_running_loop().run_in_executor(
            None,
            lambda: chain.process(chunks),
        )

        return KadiaResponse(answer=answer.strip(), chunks=chunks)

    async def chat_science(
        self, query: str, n_chunks: int, n_documents: int, minimum_score: float = 0.5
    ) -> KadiaResponse:
        """
        Flow for searching relevant document for the query and then answering query using documents and LLM

        :param query:
        :param n_chunks:
        :param n_documents:
        :param minimum_score:
        :return:
        """
        if n_chunks:
            scored_chunks = await self.semantic_search(
                query=query,
                n_chunks=n_chunks,
                n_documents=n_documents,
                minimum_score=minimum_score,
            )
            chunks = [scored_chunk.chunk for scored_chunk in scored_chunks]

            chain = QAChain(query=query, llm_manager=self.model.llm_manager)
            answer = await asyncio.get_running_loop().run_in_executor(
                None,
                lambda: chain.process(chunks),
            )
            return KadiaResponse(answer=answer.strip(), chunks=chunks)
        else:
            answer = await asyncio.get_running_loop().run_in_executor(
                None,
                lambda: self.model.llm_manager.process(
                    self.model.llm_manager.prompter.question(query)
                ),
            )
            return KadiaResponse(answer=answer.strip(), chunks=[])

    async def summarize_document(self, document_query):
        document_id = await self.upsert_document_by_query(
            document_query, skip_downloading_pdf=False
        )
        chunks = await asyncio.get_running_loop().run_in_executor(
            None,
            lambda: self.vector_storage.get_by_field_values(
                field_values=(("document_id", document_id),)
            ),
        )
        chain = SummarizeChain(llm_manager=self.model.llm_manager)
        answer = await asyncio.get_running_loop().run_in_executor(
            None,
            lambda: chain.process(chunks),
        )
        return KadiaResponse(answer=answer.strip(), chunks=chunks)

    async def general_text_processing(self, request, text):
        """
        Process user's request using text

        :param request:
        :param text:
        :return:
        """
        answer = await asyncio.get_running_loop().run_in_executor(
            None,
            lambda: self.model.llm_manager.process(
                self.model.llm_manager.prompter.general_text_processing(
                    request=request, text=text
                )
            ),
        )
        return KadiaResponse(answer=answer.strip(), chunks=[])



from typing import (
    Iterable,
    List,
    Optional,
    Tuple,
)


class BaseVectorStorage:
    def query(
        self,
        query_embedding: List[float],
        n_chunks: int,
        field_values: Optional[Iterable[Tuple[str, str]]] = None,
    ):
        raise NotImplementedError()



import confuse
import torch
from ctransformers import AutoModelForCausalLM
from keybert import KeyBERT
from langchain import OpenAI
from lazy import lazy
from transformers import AutoTokenizer

from .llm_manager import LLMManager
from .prompts.base import BasePrompter


class KadiaModel:
    """
    Utility class that manages all nested AI models required for Kadia to be functional.
    Mainly consists of configs and models instances.
    """

    def __init__(self, config: confuse.Configuration):
        self.config = config

    @lazy
    def keyword_extractor(self):
        if self.config["keyword_extraction"]:
            return KeyBERT()

    @lazy
    def llm_manager(self):
        if self.config["llm"]["model_type"] == "llama":
            return LLMManager(
                llm=AutoModelForCausalLM.from_pretrained(
                    **self.config["llm"]["config"].get(dict)
                ),
                prompter=BasePrompter.prompter_from_type(
                    self.config["llm"]["prompter"]["type"].get(str)
                ),
                config=self.config["llm"]["config"].get(dict),
                max_prompt_chars=self.config["llm"]["max_prompt_chars"].get(int),
            )
        elif self.config["llm"]["model_type"] == "mistral":
            return LLMManager(
                llm=AutoModelForCausalLM.from_pretrained(
                    **self.config["llm"]["config"].get(dict)
                ),
                prompter=BasePrompter.prompter_from_type(
                    self.config["llm"]["prompter"]["type"].get(str)
                ),
                config=self.config["llm"]["config"],
                max_prompt_chars=self.config["llm"]["max_prompt_chars"].get(int),
            )
        elif self.config["llm"]["model_type"] == "openai":
            return LLMManager(
                llm=OpenAI(**self.config["llm"]["config"].get(dict)),
                prompter=BasePrompter.prompter_from_type(
                    self.config["llm"]["prompter"]["type"].get(str)
                ),
                config=self.config["llm"]["config"],
                max_prompt_chars=self.config["llm"]["max_prompt_chars"].get(int),
            )
        elif self.config["llm"]["model_type"].get(str) == "petals":
            from petals import AutoDistributedModelForCausalLM

            return LLMManager(
                llm=AutoDistributedModelForCausalLM.from_pretrained(
                    self.config["llm"]["config"]["model_name"].get(str),
                    torch_dtype=getattr(
                        torch, self.config["llm"]["config"]["torch_dtype"].get(str)
                    ),
                    low_cpu_mem_usage=True,
                ),
                prompter=BasePrompter.prompter_from_type(
                    self.config["llm"]["prompter"]["type"].get(str)
                ),
                config=self.config["llm"]["config"],
                max_prompt_chars=self.config["llm"]["max_prompt_chars"].get(int),
                tokenizer=AutoTokenizer.from_pretrained(
                    self.config["llm"]["config"]["model_name"].get(str)
                ),
            )



class ConfigGenerator:
    @classmethod
    def default_config(
        cls,
        llm_name: str = "mistral-7b",
        embedder_name: str = "bge-small-en",
        device: str = "cpu",
        gpu_layers: int = 50,
    ):
        llm_config = cls.standard_llms(llm_name)
        if device == "gpu" or "cuda" in device:
            llm_config["config"]["gpu_layers"] = gpu_layers
        return {
            "text_splitter": {
                "add_metadata": True,
                "chunk_size": 1024,
                "chunk_overlap": 128,
            },
            "embedder": cls.standard_embedders(embedder_name, device=device),
            "keyword_extraction": True,
            "llm": llm_config,
        }

    @classmethod
    def standard_embedders(cls, name, device: str = "cpu"):
        return {
            "instructor-xl": {
                "model_name": "hkunlp/instructor-xl",
                "model_kwargs": {"device": device},
                "model_type": "instructor",
            },
            "instructor-large": {
                "model_name": "hkunlp/instructor-large",
                "model_kwargs": {"device": device},
                "model_type": "instructor",
            },
            "bge-small-en": {
                "model_name": "BAAI/bge-small-en-v1.5",
                "model_kwargs": {"device": device},
                "model_type": "bge",
            },
            "bge-base-en": {
                "model_name": "BAAI/bge-base-en-v1.5",
                "model_kwargs": {"device": device},
                "model_type": "bge",
            },
            "bge-large-en": {
                "model_name": "BAAI/bge-large-en-v1.5",
                "model_kwargs": {"device": device},
                "model_type": "bge",
            },
            "openai": {"model_name": "text-embedding-ada-002", "model_type": "openai"},
        }[name]

    @classmethod
    def standard_llms(cls, name):
        return {
            "mistral-7b": {
                "config": {
                    "context_length": 8192,
                    "max_new_tokens": 1024,
                    "model_path_or_repo_id": "TheBloke/Mistral-7B-Instruct-v0.1-GGUF",
                    "model_file": "mistral-7b-instruct-v0.1.Q5_K_M.gguf",
                },
                "max_prompt_chars": int(8192 * 2.5),
                "model_type": "mistral",
                "prompter": {"type": "mistral-7b"},
            },
            "llama-2-7b": {
                "config": {
                    "context_length": 4096,
                    "max_new_tokens": 512,
                    "model_path_or_repo_id": "TheBloke/Llama-2-7B-Chat-GGML",
                    "model_file": "llama-2-7b-chat.ggmlv3.q4_K_S.bin",
                },
                "max_prompt_chars": int(4096 * 2.5),
                "model_type": "llama",
                "prompter": {"type": "llama-7b"},
            },
            "llama-2-7b-uncensored": {
                "config": {
                    "context_length": 4096,
                    "max_new_tokens": 512,
                    "model_path_or_repo_id": "TheBloke/Luna-AI-Llama2-Uncensored-GGML",
                    "model_file": "luna-ai-llama2-uncensored.ggmlv3.q4_K_S.bin",
                },
                "max_prompt_chars": int(4096 * 2.5),
                "model_type": "llama",
                "prompter": {"type": "default"},
            },
            "llama-2-13b": {
                "config": {
                    "context_length": 4096,
                    "max_new_tokens": 512,
                    "model_file": "llama-2-13b-chat.ggmlv3.q4_K_S.bin",
                    "model_path_or_repo_id": "TheBloke/Llama-2-13B-chat-GGML",
                },
                "max_prompt_chars": int(4096 * 2.5),
                "model_type": "llama",
                "prompter": {"type": "llama-7b"},
            },
            "petals-llama-2-70b": {
                "config": {
                    "max_new_tokens": 512,
                    "model_name": "meta-llama/Llama-2-7b-chat-hf",
                    "torch_dtype": "float32",
                },
                "max_prompt_chars": int(8192 * 2.5),
                "model_type": "petals",
                "prompter": {"type": "llama-7b"},
            },
            "petals-stable-beluga": {
                "config": {
                    "max_new_tokens": 512,
                    "model_name": "stabilityai/StableBeluga2",
                    "torch_dtype": "float32",
                },
                "max_prompt_chars": int(8192 * 2.5),
                "model_type": "petals",
                "prompter": {"type": "beluga"},
            },
            "openai": {
                "config": {},
                "max_prompt_chars": int(4096 * 3.5),
                "model_type": "openai",
                "prompter": {"type": "default"},
            },
        }[name]



import logging
from dataclasses import dataclass
from datetime import datetime
from typing import (
    List,
)

from izihawa_textutils.utils import remove_markdown
from stc_geck.advices import BaseDocumentHolder
from unstructured.cleaners.core import clean

from kadia.text_splitters import (
    MarkdownHeaderTextSplitter,
    RecursiveCharacterTextSplitter,
)

BANNED_SECTIONS = {
    "author contribution",
    "data availability statement",
    "declaration of competing interest",
    "acknowledgments",
    "acknowledgements",
    "supporting information",
    "conflict of interest disclosures",
    "conflict of interest",
    "conflict of interest statement",
    "ethics statement",
    "references",
    "external links",
    "further reading",
    "works cited",
    "bibliography",
    "notes",
    "sources",
    "footnotes",
    "suggested readings",
}


@dataclass
class Chunk:
    document_id: str
    field: str
    chunk_id: int
    title: str
    text: str
    start_index: int
    length: int


@dataclass
class StoredChunk:
    document_id: str
    field: str
    chunk_id: int
    title: str
    start_index: int
    length: int


def extract_title_parts(document_holder, split):
    title_parts = [document_holder.title]
    for hn in range(6):
        if hn_value := split.metadata.get(f"h{hn}"):
            title_parts.append(hn_value)
    return title_parts


class DocumentChunker:
    def __init__(
        self,
        chunk_size: int = 1024,
        chunk_overlap: int = 128,
        add_metadata: bool = False,
        add_year: bool = True,
    ):
        self.text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            keep_separator=True,
            strip_whitespace=True,
            length_function=lambda text: len(remove_markdown(text)),
        )
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.add_metadata = add_metadata
        self.add_year = add_year

    def _splits_to_chunks(
        self, document_holder: BaseDocumentHolder, field: str, splits
    ):
        chunks = []
        for chunk_id, split in enumerate(splits):
            page_content = str(split.page_content)
            chunk_text = clean(
                remove_markdown(page_content),
                extra_whitespace=True,
                dashes=True,
                bullets=True,
                trailing_punctuation=True,
            )
            parts = [chunk_text]
            title_parts = extract_title_parts(document_holder, split)
            if len(chunk_text) < 128 or any(
                [title_part.lower() == "references" for title_part in title_parts]
            ):
                continue
            if self.add_metadata:
                parts.append(f'TITLE: {" ".join(title_parts)}')
                if document_holder.has_field("keywords"):
                    keywords = ", ".join(document_holder.keywords)
                    parts.append(f"KEYWORDS: {keywords}")
                if document_holder.has_field("tags"):
                    tags = ", ".join(document_holder.tags)
                    parts.append(f"TAGS: {tags}")
            if self.add_year and document_holder.has_field("issued_at"):
                issued_at = datetime.utcfromtimestamp(document_holder.issued_at)
                parts.append(f"YEAR: {issued_at.year}")
            text = "\n".join(parts)
            chunks.append(
                Chunk(
                    document_id=document_holder.get_id(),
                    field=field,
                    chunk_id=chunk_id,
                    title="\n".join(title_parts),
                    text=text,
                    start_index=split.metadata["start_index"],
                    length=len(page_content),
                )
            )
        return chunks

    def to_chunks(self, document_holder: BaseDocumentHolder) -> List[Chunk]:
        logging.getLogger("statbox").info(
            {
                "action": "chunking",
                "document_id": document_holder.get_id(),
                "mode": "kadia",
            }
        )
        document = document_holder.document
        abstract = document.get("abstract", "")
        content = document.get("content", "")

        headers_to_split_on = [
            ("#", "h1"),
            ("##", "h2"),
            ("###", "h3"),
            ("####", "h4"),
            ("#####", "h5"),
            ("######", "h6"),
        ]

        # MD splits
        markdown_splitter = MarkdownHeaderTextSplitter(
            headers_to_split_on=headers_to_split_on
        )
        md_abstract_header_splits = markdown_splitter.split_text(abstract)
        md_content_header_splits = markdown_splitter.split_text(content)

        # Split
        abstract_splits = self.text_splitter.split_documents(md_abstract_header_splits)
        content_splits = self.text_splitter.split_documents(md_content_header_splits)

        chunks = []
        chunks.extend(
            self._splits_to_chunks(document_holder, "abstract", abstract_splits)
        )
        chunks.extend(
            self._splits_to_chunks(document_holder, "content", content_splits)
        )

        return chunks



import functools
import logging
import re
import sys
import textwrap
from typing import Optional

import fire
from termcolor import colored

from .kadia_ai import KadiaAI
from .exceptions import QdrantStorageNotAvailableError


def exception_handler(func):
    @functools.wraps(func)
    async def wrapper_func(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except QdrantStorageNotAvailableError as e:
            print(
                f"{colored('INFO', 'red')}: Cannot connect to Qdrant: {e.info}\n"
                f"{colored('HINT', 'yellow')}: Launch qdrant using `docker run -p 6333:6333 -p 6334:6334 qdrant/qdrant`",
                file=sys.stderr,
            )
            raise

    return wrapper_func


class KadiaCli:
    def __init__(self, kadia: Optional[KadiaAI] = None):
        self.kadia = kadia or KadiaAI()

    async def add_all_documents(self, query: str = "", batch_size: int = 200):
        async with self.kadia as kadia:
            batch = []
            async for document in kadia.data_source.stream_documents(query=query):
                logging.getLogger("statbox").info(
                    {"action": "collect_document", "document": document.document_id}
                )
                batch.append(document)
                if len(batch) >= batch_size:
                    await self.kadia.upsert_documents(batch)
                    batch = []
            if batch:
                await self.kadia.upsert_documents(batch)

    @exception_handler
    async def chat_doc(
        self,
        document_query: str,
        query: str,
        n_chunks: int = 5,
        minimum_score: float = 0.5,
    ):
        """
        Ask a question about content of document identified by DOI.

        :param document_query: query that returns unique document
        :param query: Text query to the document
        :param n_chunks: the number of chunks to extract
            more means more tokens to use and more precision in answer
        :param minimum_score:
        """
        async with self.kadia as kadia:
            print(f"{colored('Document', 'green')}: {document_query}")
            print(f"{colored('Q', 'green')}: {query}")
            kadia_response = await kadia.chat_document(
                document_query,
                query,
                n_chunks,
                minimum_score=minimum_score,
            )
            print(f"{colored('A', 'green')}: {kadia_response.answer}")

    @exception_handler
    async def chat_sci(
        self,
        query: str,
        n_chunks: int = 5,
        n_documents: int = 10,
        minimum_score: float = 0.5,
    ):
        """
        Ask a general questions

        :param query: text query to the document
        :param n_chunks: the number of chunks to extract
            more means more tokens to use and more precision in answer
        :param n_documents: the number of chunks to extract
            more means more tokens to use and more precision in answer
        :param minimum_score:
        """
        async with self.kadia as kadia:
            print(f"{colored('Q', 'green')}: {query}")
            kadia_response = await kadia.chat_science(
                query=query,
                n_chunks=n_chunks,
                n_documents=n_documents,
                minimum_score=minimum_score,
            )
            answer = re.sub(
                r"\(DOI:\s*([^)]+)\)",
                r"(https://doi.org/\g<1>)",
                kadia_response.answer,
            )
            references = []
            visited = set()
            for chunk in kadia_response.chunks:
                field, value = chunk.document_id.split(":", 1)
                document_id = f"{field}:{value}"
                if document_id in visited:
                    continue
                visited.add(document_id)
                title = chunk.title.replace("\n", " - ")
                references.append(f"{document_id}: {title}")
            references_str = "\n".join(references)
            print(f"{colored('A', 'green')}: {answer}")
            print(
                f"{colored('References', 'green')}:\n{textwrap.indent(references_str, ' - ')}"
            )

    @exception_handler
    async def chunk_document(
        self,
        document_query: str,
    ):
        """
        Ask a question about content of document identified by DOI.

        :param document_query: query that returns unique document
        """
        async with self.kadia as kadia:
            print(f"{colored('Document', 'green')}: {document_query}")
            documents = await kadia.data_source.search_documents(
                document_query, limit=1
            )
            document_chunks = await kadia.generate_chunks_from_document(documents[0])
            print(f"{colored('Chunks', 'green')}: {document_chunks}")

    @exception_handler
    async def sum_doc(self, document_query: str):
        """
        Summarization of the document

        :param document_query: query that returns unique document
        """
        async with self.kadia as kadia:
            print(f"{colored('Document', 'green')}: {document_query}")
            kadia_response = await kadia.summarize_document(document_query)
            print(f"{colored('Summarization', 'green')}: {kadia_response.answer}")

    @exception_handler
    async def semantic_search(
        self,
        query: str,
        n_chunks: int = 5,
        n_documents: int = 10,
        minimum_score: float = 0.5,
        use_only_keywords: bool = True,
    ):
        """
        Search related to query text chunks among `n` documents

        :param query: query to STC
        :param n_chunks: number of chunks to return
        :param n_documents: the number of documents to extract from STC
        :param minimum_score:
        :param use_only_keywords:
        """
        async with self.kadia as kadia:
            print(f"{colored('Q', 'green')}: {query}")
            scored_chunks = await kadia.semantic_search(
                query=query,
                n_chunks=n_chunks,
                n_documents=n_documents,
                minimum_score=minimum_score,
                use_only_keywords=use_only_keywords,
            )
            references = []
            for scored_chunk in scored_chunks:
                field, value = scored_chunk.chunk.document_id.split(":", 1)
                document_id = f"{field}:{value}"
                title = scored_chunk.chunk.title.replace("\n", " - ")
                references.append(
                    f" - {document_id}: {title}\n   {scored_chunk.chunk.text}"
                )
            references = "\n".join(references)
            print(f"{colored('References', 'green')}:\n{references}")


def kadia_cli(debug: bool = False):
    """
    :param debug: add debugging output
    :return:
    """
    logging.basicConfig(
        stream=sys.stdout, level=logging.INFO if debug else logging.ERROR
    )
    kadia = KadiaCli()
    return {
        "add-all-documents": kadia.add_all_documents,
        "chat-doc": kadia.chat_doc,
        "chat-sci": kadia.chat_sci,
        "chunk_document": kadia.chunk_document,
        "semantic-search": kadia.semantic_search,
        "sum-doc": kadia.sum_doc,
        "write-config": kadia.kadia.ensure_config,
    }


def run():
    fire.Fire(kadia_cli, name="kadia")


if __name__ == "__main__":
    run()



from typing import List

from langchain.embeddings import (
    HuggingFaceBgeEmbeddings,
    HuggingFaceInstructEmbeddings,
    OpenAIEmbeddings,
)


class Embedder:
    """
    Utility class that manages all nested AI models required for Kadia to be functional.
    Mainly consists of configs and models instances.
    """

    def __init__(
        self, model_type: str, model_name: str, model_kwargs: dict | None = None
    ):
        self.model_name = model_name
        match model_type:
            case "instructor":
                self._embedder = HuggingFaceInstructEmbeddings(
                    model_name=model_name,
                    model_kwargs=model_kwargs,
                    embed_instruction="Represent science paragraph for retrieval",
                    query_instruction="Represent science question for retrieval",
                )
            case "bge":
                self._embedder = HuggingFaceBgeEmbeddings(
                    model_name=model_name,
                    model_kwargs=model_kwargs,
                    encode_kwargs={"normalize_embeddings": True},
                )
            case "openai":
                self._embedder = OpenAIEmbeddings(model=model_name)
            case _:
                raise ValueError("Unsupported embedding model")

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """
        Embed texts into vectors using selected models

        :param texts: a list of texts
        :return: list of vectors
        """
        return self._embedder.embed_documents(texts)

    def get_embeddings_id(self):
        return self.model_name.replace("/", "-")



from dataclasses import dataclass
from typing import (
    List,
    Optional,
    AsyncGenerator,
    AsyncIterator,
)


@dataclass
class SourceDocument:
    document: dict
    document_id: str


class BaseDataSource:
    def stream_documents(
        self,
        query: str,
        limit: int = 0,
    ) -> AsyncIterator[SourceDocument]:
        raise NotImplementedError()

    async def search_documents(
        self,
        query: str,
        limit: int = 5,
    ) -> List[SourceDocument]:
        raise NotImplementedError()



import logging

import confuse


class LLMManager:
    def __init__(
        self,
        llm,
        prompter,
        config: confuse.Configuration,
        max_prompt_chars: int,
        tokenizer=None,
    ):
        self.llm = llm
        self.prompter = prompter
        self.config = config
        self.max_prompt_chars = max_prompt_chars
        self.tokenizer = tokenizer

    @property
    def context_length(self):
        return self.config["context_length"].get(int)

    def process(self, prompt):
        logging.getLogger("statbox").info(
            {"action": "process", "mode": "llm_manager", "prompt": prompt}
        )
        if self.tokenizer:
            input_ids = self.tokenizer(prompt, return_tensors="pt")["input_ids"]
            outputs = self.llm.generate(
                input_ids,
                max_new_tokens=self.config["max_new_tokens"].get(int),
                temperature=self.config["temperature"].get(float),
            )
            return self.tokenizer.batch_decode(outputs[:, input_ids.shape[1] :])[
                0
            ].replace("</s>", "")
        else:
            return self.llm(prompt)



from typing import (
    List,
    Optional,
    AsyncGenerator,
    AsyncIterator,
)

import orjson
from stc_geck.advices import BaseDocumentHolder
from stc_geck.client import StcGeck

from .base import (
    BaseDataSource,
    SourceDocument,
)


class GeckDataSource(BaseDataSource):
    def __init__(self, geck: StcGeck):
        self.geck = geck

    def _query_function(
        self,
        query: str = "",
        with_language_filter: bool = True,
        with_type_filter: bool = True,
        with_existence_filter: bool = False,
    ):
        subqueries = []
        if with_type_filter:
            subqueries.append(
                {
                    "occur": "must",
                    "query": {
                        "boolean": {
                            "subqueries": [
                                {
                                    "occur": "should",
                                    "query": {
                                        "term": {"field": "type", "value": "book"}
                                    },
                                },
                                {
                                    "occur": "should",
                                    "query": {
                                        "term": {
                                            "field": "type",
                                            "value": "edited-book",
                                        }
                                    },
                                },
                                {
                                    "occur": "should",
                                    "query": {
                                        "term": {"field": "type", "value": "monograph"}
                                    },
                                },
                                {
                                    "occur": "should",
                                    "query": {
                                        "term": {
                                            "field": "type",
                                            "value": "reference-book",
                                        }
                                    },
                                },
                                {
                                    "occur": "should",
                                    "query": {
                                        "term": {
                                            "field": "type",
                                            "value": "journal-article",
                                        }
                                    },
                                },
                                {
                                    "occur": "should",
                                    "query": {
                                        "term": {"field": "type", "value": "wiki"}
                                    },
                                },
                            ]
                        }
                    },
                }
            )
        if with_language_filter:
            subqueries.append(
                {
                    "occur": "must",
                    "query": {"term": {"field": "languages", "value": "en"}},
                }
            )
        if with_existence_filter:
            subqueries.append(
                {"occur": "must", "query": {"exists": {"field": "content"}}}
            )
        if query:
            subqueries.append(
                {"occur": "must", "query": {"match": {"value": query.lower()}}}
            )
        if subqueries:
            return {"boolean": {"subqueries": subqueries}}
        else:
            return {"all": {}}

    async def stream_documents(
        self,
        query: str,
        limit: int = 0,
    ) -> AsyncIterator[SourceDocument]:
        documents = self.geck.get_summa_client().documents(
            self.geck.index_alias,
            query_filter=self._query_function(query, with_existence_filter=True),
        )
        counter = 0
        async for document in documents:
            document = orjson.loads(document)
            document_holder = BaseDocumentHolder(document)
            document_id = document_holder.get_id()
            if not document_id:
                continue
            yield SourceDocument(
                document=document,
                document_id=document_id,
            )
            counter += 1
            if limit and counter >= limit:
                return

    async def search_documents(
        self,
        query: str,
        limit: int = 5,
    ) -> List[SourceDocument]:
        documents = await self.geck.get_summa_client().search_documents(
            {
                "index_alias": self.geck.index_alias,
                "query": self._query_function(query),
                "collectors": [{"top_docs": {"limit": limit}}],
                "is_fieldnorms_scoring_enabled": False,
            }
        )
        source_documents = []
        for document in documents:
            document_holder = BaseDocumentHolder(document)
            document_id = document_holder.get_id()
            if not document_id:
                continue
            source_documents.append(
                SourceDocument(
                    document=document,
                    document_id=document_id,
                )
            )
        return source_documents



import asyncio
import dataclasses
import hashlib
import logging
from typing import (
    Iterable,
    List,
    Optional,
    Tuple,
)

import grpc
from aiokit import MultipleAsyncExecution
from qdrant_client import QdrantClient
from qdrant_client.models import (
    Distance,
    FieldCondition,
    Filter,
    MatchValue,
    PayloadSchemaType,
    PointStruct,
    Range,
    VectorParams,
)
from stc_geck.advices import BaseDocumentHolder

from ..document_chunker import Chunk, DocumentChunker, StoredChunk
from ..exceptions import QdrantStorageNotAvailableError
from .base import BaseVectorStorage


@dataclasses.dataclass
class ScoredChunk:
    chunk: StoredChunk
    score: float


class QdrantVectorStorage(BaseVectorStorage):
    def __init__(
        self,
        qdrant_config: dict,
        collection_name,
        force_recreate: bool = False,
    ):
        self.db = QdrantClient(**qdrant_config)
        self.collection_name = collection_name
        self.force_recreate = force_recreate
        self.is_existing = False

    def _exists_collection(self, collection_name):
        if self.is_existing:
            return True
        try:
            collections = self.db.get_collections()
        except grpc.RpcError:
            raise QdrantStorageNotAvailableError()
        for collection in collections.collections:
            if collection.name == collection_name:
                self.is_existing = True
                return True
        return False

    def _ensure_collection(self, collection_name, size):
        if not self.force_recreate:
            if self._exists_collection(collection_name):
                return
            self.db.create_collection(
                collection_name=collection_name,
                vectors_config=VectorParams(size=size, distance=Distance.COSINE),
            )
        else:
            self.db.recreate_collection(
                collection_name=collection_name,
                vectors_config=VectorParams(size=size, distance=Distance.COSINE),
            )
            self.force_recreate = False
        self.is_existing = True
        self.db.create_payload_index(
            collection_name=collection_name,
            field_name="document_id",
            field_schema=PayloadSchemaType.KEYWORD,
        )

    def get_by_field_values(
        self, field_values: Iterable[Tuple[str, str]], sort: bool = True
    ) -> List[Chunk]:
        if not self._exists_collection(self.collection_name):
            return []
        points, _ = self.db.scroll(
            collection_name=self.collection_name,
            scroll_filter=Filter(
                should=[
                    FieldCondition(key=field, match=MatchValue(value=value))
                    for (field, value) in field_values
                ],
            ),
            limit=2**31,
        )
        payloads = [Chunk(**point.payload) for point in points]
        if sort:
            return list(sorted(payloads, key=lambda x: (x.document_id, x.chunk_id)))
        return payloads

    def exists_by_field_value(self, field, value) -> bool:
        if not self._exists_collection(self.collection_name):
            return False
        points, _ = self.db.scroll(
            collection_name=self.collection_name,
            scroll_filter=Filter(
                should=[
                    FieldCondition(key=field, match=MatchValue(value=value)),
                ],
            ),
            with_payload=False,
            limit=1,
        )
        return len(points) > 0

    def query(
        self,
        query_embedding,
        n_chunks: int,
        field_values: Optional[Iterable[Tuple[str, str]]] = None,
    ) -> List[ScoredChunk]:
        self._ensure_collection(
            collection_name=self.collection_name,
            size=len(query_embedding),
        )
        if n_chunks == 0:
            return []
        query_filter = None
        if field_values:
            conditions = []
            for (field, value) in field_values:
                if isinstance(value, str):
                    conditions.append(
                        FieldCondition(key=field, match=MatchValue(value=value))
                    )
                else:
                    conditions.append(
                        FieldCondition(
                            key=field,
                            range=Range(
                                gte=value,
                                lte=value,
                            ),
                        )
                    )
            query_filter = Filter(should=conditions)
        points = self.db.search(
            collection_name=self.collection_name,
            query_vector=query_embedding,
            query_filter=query_filter,
            limit=n_chunks,
        )
        return [
            ScoredChunk(chunk=StoredChunk(**point.payload), score=point.score)
            for point in points
        ]

    def upsert(self, chunks: list[Chunk], embeddings: list[list[float]]):
        if not chunks:
            return
        embedding_size = len(embeddings[0])
        self._ensure_collection(
            collection_name=self.collection_name,
            size=embedding_size,
        )
        return self.db.upsert(
            collection_name=self.collection_name,
            points=[
                PointStruct(
                    id=hashlib.md5(
                        f"{chunk.document_id}@{chunk.chunk_id}".encode()
                    ).hexdigest(),
                    vector=embedding,
                    payload=dataclasses.asdict(
                        StoredChunk(
                            document_id=chunk.document_id,
                            field=chunk.field,
                            chunk_id=chunk.chunk_id,
                            title=chunk.title,
                            start_index=chunk.start_index,
                            length=chunk.length,
                        )
                    ),
                )
                for chunk, embedding in zip(chunks, embeddings)
            ],
        )

    async def _to_chunks(
        self,
        document_holder: BaseDocumentHolder,
        document_chunker: DocumentChunker,
        all_chunks: list[Chunk],
    ):
        all_chunks.extend(
            await asyncio.get_running_loop().run_in_executor(
                None, lambda d: document_chunker.to_chunks(d), document_holder
            )
        )

    async def get_missing_chunks(
        self,
        document_holders: List[BaseDocumentHolder],
        document_chunker: DocumentChunker,
        par: int = 16,
    ) -> List[Chunk]:
        all_chunks = []
        executor = MultipleAsyncExecution(par)
        for document_holder in document_holders:
            is_stored = await asyncio.get_running_loop().run_in_executor(
                None,
                lambda document_id: self.exists_by_field_value(
                    "document_id", document_id
                ),
                document_holder.get_id(),
            )
            if is_stored:
                logging.getLogger("statbox").info(
                    {
                        "action": "already_stored",
                        "mode": "kadia",
                        "document_id": document_holder.get_id(),
                    }
                )
                continue
            if (
                "abstract" not in document_holder.document
                and "content" not in document_holder.document
            ):
                logging.getLogger("statbox").info(
                    {
                        "action": "no_content",
                        "mode": "kadia",
                        "document_id": document_holder.get_id(),
                    }
                )
                continue
            try:
                await executor.execute(
                    self._to_chunks(document_holder, document_chunker, all_chunks)
                )
            except ValueError:
                logging.getLogger("statbox").info(
                    {
                        "action": "broken_content",
                        "mode": "kadia",
                        "document_id": document_holder.get_id(),
                    }
                )
        await executor.join()
        return all_chunks
